;;;;
;;;; b i n t r e e . s c m	     -- binary trees, as used to represent
;;;;					lists, sets, etc.
;;;;

(provide 'bintree)
(require 'vectree)


;;;
;;; Basic constructors and destruct(urat)ors
;;;
(define bt:node-tag (make-tag 'B2))
(define (bt:node x y) (vector bt:node-tag x y))
(define bt:node? (vectree-detector bt:node-tag 2))
(define bt:left  v-1st)
(define bt:right v-2nd)
(define bt:set-left!  v-set-1st!)
(define bt:set-right! v-set-2nd!)

(define bt:leaf-tag (make-tag 'B1))
(define (bt:leaf x) (vector bt:leaf-tag x))
(define bt:leaf? (vectree-detector bt:leaf-tag 1))
(define bt:elem  v-1st)

(define bt:nil   '())
(define bt:null? null?)

(define (bt:new)  (bt:node bt:nil bt:nil))


;;;
;;; Easy operations
;;;
(define (bt:cons elem bt)
  (bt:node (bt:leaf elem) bt))

(define (bt:append l r)
  (cond
   ((bt:null? l) r)
   ((bt:null? r) l)
   (else (bt:node l r))))

(define (bt:reverse bt)
  (cond
   ((bt:null? bt) bt)
   ((bt:leaf? bt) bt)
   ((bt:node? bt) (bt:node (bt:right bt) (bt:left bt)))
   (else (error "not a binary tree" bt))))


;;;
;;; Simple traversal of a binary-tree
;;;
(define (bt:lookup bt fun found not-found)
    (cond
     ((bt:null? bt) (not-found))
     ((bt:leaf? bt)
        (let ((e (bt:elem bt)))
	  (if (fun e) (found e) (not-found))))
     ((bt:node? bt)
        (let*
	    ((left (bt:left bt))
	     (right (bt:right bt))
	     (lookup-right
	        (lambda ()
		  (bt:lookup right fun found not-found))))
	  (bt:lookup left fun found lookup-right)))
   (else (error "not a binary tree" bt))))

(define (bt:empty? bt)
  (bt:lookup bt (lambda (x) #t)
     (lambda (x) #t)
     (lambda () #f)))


;;;
;;; Splitting a binary tree into halves
;;;
(define (bt:do-split n accum bt cont not-enough)
  (cond
   ((zero? n)
    (cont accum bt))
   ((bt:null? bt) (not-enough n accum))
   ((bt:leaf? bt) (bt:do-split (1- n) (append accum (list (bt:elem bt)))
			       cont not-enough))
   ((bt:node? bt)
    (let*
	((left (bt:left bt))
	 (right (bt:right bt))
	 (found-left
	  (lambda (list rest)
	    (found list (bt:node rest right))))
	 (not-enough-left
	  (lambda (n accum)
	    (bt:do-split n accum right cont not-enough))))
      (bt:do-split n left fun found-left not-enough-left)))
   (else (error "not a binary tree" bt))))

(define (bt:car bt)
  (bt:lookup bt (lambda (x) #f) 
     (lambda (x) x)
     (lambda () (error "empty bt has no car"))))
(define (bt:cdr bt)
  (bt:do-split 1 '() bt
     (lambda (res bt) (car res))
     (lambda foo (error "empty bt has no cdr"))))
(define (bt:n-car n bt)
  (bt:do-split n '() bt
     (lambda (res bt) res)
     (lambda (n accum) (error "bt should be longer by" n))))
(define (bt:n-cdr n bt)
  (bt:do-split n '() bt
     (lambda (res bt) bt)
     (lambda (n accum) (error "bt should be longer by" n))))


;;;
;;; binary trees as lists
;;;
(define (bt->list bt)
  (cond
   ((bt:null? bt) bt)
   ((bt:leaf? bt) (bt:elem bt))
   ((bt:node? bt) (append (bt->list (bt:left bt)) (bt->list (bt:right bt))))
   (else (error "not a binary tree" bt))))
(define (list->bt list)
  (cond
   ((null? list) list)
   ((pair? list)
     (bt:cons (car list) (bt->list (cdr list))))
   (else (error "not a list" list))))

;;;
;;; Length
;;;
(define (bt:length bt)
  (length (bt->list bt))) ; optimizing is no *my* job!
